"""Utilities for training flows, fitting to samples or ysing variational inference."""
from .data_fit import fit_to_data
from .variational_fit import fit_to_variational_target

__all__ = [
    "fit_to_data",
    "fit_to_variational_target",
]
